import asyncio
import logging
import re

from typing import Callable

from .ble_protocol import BLEProtocol
from .on_receive import OnReceive
from .rfid import RfidCommands
from .serial_protocol import SerialProtocol
from .tcp_protocol import TCPProtocol
from .write_commands import WriteCommands

from smartx_rfid.utils.event import on_event
from smartx_rfid.devices._base import DeviceBase

ant_default_config = {
    "1": {"active": True, "power": 22, "rssi": -120},
    "2": {"active": False, "power": 22, "rssi": -120},
    "3": {"active": False, "power": 22, "rssi": -120},
    "4": {"active": False, "power": 22, "rssi": -120},
}


class X714(DeviceBase, SerialProtocol, OnReceive, RfidCommands, BLEProtocol, WriteCommands, TCPProtocol):
    """RFID reader that supports SERIAL, TCP and BLE connections."""

    def __init__(
        self,
        name: str = "X714",
        connection_type: str = "SERIAL",  # SERIAL, BLE or TCP
        # SERIAL
        port: str = "AUTO",  # AUTO or COMx / /dev/ttyX
        baudrate: int = 115200,
        vid: int = 1,
        pid: int = 1,
        # TCP
        ip: str | None = "192.168.1.100",
        tcp_port: int = 23,
        # BLE
        ble_name: str = "SMTX",
        # GENERIC
        buzzer: bool = False,
        session: int = 1,  # 0, 1, 2, 3
        start_reading: bool = False,
        gpi_start: bool = False,
        always_send: bool = True,
        simple_send: bool = False,
        keyboard: bool = False,
        decode_gtin: bool = False,
        hotspot: bool = True,
        reconnection_time: int = 3,
        prefix: str = "",
        protected_inventory_active: bool = False,
        protected_inventory_password: str | None = "12345678",
        # Antenna config
        # If ant_dict is provided use it else use the other vars
        ant_dict: dict | None = None,
        active_ant: list[int] | None = [1],
        read_power: int = 22,
        read_rssi: int = -120,
        **kwargs,
    ):
        """
        Create X714 RFID reader.

        Args:
            name: Device name
            connection_type: How to connect (SERIAL, TCP, BLE)
            port: Serial port or AUTO to detect
            baudrate: Serial speed
            vid: USB vendor ID for auto-detect
            pid: USB product ID for auto-detect
            ip: IP address for TCP connection
            tcp_port: TCP port number
            ble_name: Bluetooth device name
            buzzer: Make sound when reading tags
            session: EPC session number (0-3)
            start_reading: Start reading tags automatically
            gpi_start: Use GPI input to start reading
            always_send: Send all tag events
            simple_send: Use simple tag format
            keyboard: Act like keyboard input
            decode_gtin: Decode GTIN barcodes
            hotspot: Enable hotspot mode
            reconnection_time: Seconds to wait before reconnect
            prefix: Text to add before tag data
            protected_inventory_active: Enable protected reading
            protected_inventory_password: Password for protected reading
            ant_dict: Custom antenna settings
            active_ant: Which antennas to use
            read_power: TX power in dBm
            read_rssi: RSSI threshold in dBm
        """
        DeviceBase.__init__(self)

        # Name
        self.name = name
        self.device_type = "rfid"

        # CONNECTION TYPE
        if connection_type in ["SERIAL", "BLE", "TCP"]:
            self.connection_type = connection_type
        else:
            self.connection_type = "SERIAL"
            logging.warning(f"[{self.name}] Invalid connection_type '{connection_type}' set to 'SERIAL'")

        # SERIAL CONFIG
        if not isinstance(port, str):
            port = "AUTO"
        self.port = port
        if not isinstance(baudrate, int):
            baudrate = 115200
        self.baudrate = baudrate
        if not isinstance(vid, int):
            vid = 1
        if not isinstance(pid, int):
            pid = 1
        self.vid = vid
        self.pid = pid
        self.is_auto = self.port == "AUTO"

        # TCP CONFIG
        if not isinstance(ip, str):
            ip = "192.168.1.100"
        if not isinstance(tcp_port, int):
            tcp_port = 23
        self.ip = ip
        self.tcp_port = tcp_port

        # BLE CONFIG
        if not isinstance(ble_name, str):
            ble_name = "SMTX"
        self.ble_name = ble_name
        self.init_ble_vars()

        # GENERIC CONFIG
        if not isinstance(buzzer, bool):
            buzzer = False
        self.buzzer = buzzer
        if session not in [0, 1, 2, 3]:
            session = 1
            logging.warning(f"[{self.name}] Invalid session '{session}' set to '1'")
        self.session = session

        # Modes
        if not isinstance(start_reading, bool):
            start_reading = False
        if not isinstance(gpi_start, bool):
            gpi_start = False
        if not isinstance(always_send, bool):
            always_send = True
        if not isinstance(simple_send, bool):
            simple_send = False
        if not isinstance(keyboard, bool):
            keyboard = False
        if not isinstance(decode_gtin, bool):
            decode_gtin = False
        if not isinstance(hotspot, bool):
            hotspot = True
        if not isinstance(reconnection_time, int):
            reconnection_time = 5
        self.start_reading = start_reading
        self.gpi_start = gpi_start
        self.always_send = always_send
        self.simple_send = simple_send
        self.keyboard = keyboard
        self.decode_gtin = decode_gtin
        self.hotspot = hotspot
        self.reconnection_time = reconnection_time

        # validate prefix
        if not isinstance(prefix, str):
            prefix = ""
        # validate hex
        if not re.match(r"^[0-9A-Fa-f]*$", prefix):
            prefix = ""

        self.prefix = prefix

        # PROTECTED
        if not isinstance(protected_inventory_active, bool):
            protected_inventory_active = False
        if not isinstance(protected_inventory_password, str):
            protected_inventory_password = ""
        self.protected_inventory_active = protected_inventory_active
        self.protected_inventory_password = protected_inventory_password

        # ANTENNA CONFIG
        if ant_dict is not None:
            self.ant_dict = ant_dict
        else:
            self.ant_dict = ant_default_config
            for ant in self.ant_dict.keys():
                ant_num = int(ant)
                if active_ant and ant_num in active_ant:
                    self.ant_dict[ant]["active"] = True
                else:
                    self.ant_dict[ant]["active"] = False
                self.ant_dict[ant]["power"] = read_power
                self.ant_dict[ant]["rssi"] = read_rssi

        self.serial_number = None

        self.transport = None
        self.on_con_lost = None
        self.rx_buffer = bytearray()
        self.last_byte_time = None

        self.is_connected = False
        self.is_reading = False
        self.is_gpi_trigger_on = self.gpi_start
        if self.is_gpi_trigger_on:
            self.start_reading = False

        self.on_event: Callable = on_event

    def write(self, to_send, verbose=True):
        """Send data to reader using current connection type.

        Args:
            to_send: Data to send
            verbose: Show sent data in logs
        """
        if self.connection_type == "SERIAL":
            self.write_serial(to_send, verbose)
        elif self.connection_type == "BLE":
            if hasattr(self, "create_task"):
                self.create_task(self.write_ble(to_send.encode(), verbose))
            else:
                asyncio.create_task(self.write_ble(to_send.encode(), verbose))
        else:
            if hasattr(self, "create_task"):
                self.create_task(self.write_tcp(to_send, verbose))
            else:
                asyncio.create_task(self.write_tcp(to_send, verbose))

    async def connect(self):
        """Connect to reader using configured connection type."""
        if self.connection_type == "SERIAL":
            await self.connect_serial()
        elif self.connection_type == "BLE":
            await self.connect_ble()
        else:
            await self.connect_tcp(self.ip, self.tcp_port)

    def on_connected(self):
        """Called when connection is established. Sets up reader."""
        self.config_reader()
        self.on_event(self.name, "connected", True)

    async def close(self):
        """Close connections and cancel background tasks for X714."""
        self._running = False

        # try to close underlying transport/writer depending on connection type
        try:
            if self.connection_type == "SERIAL":
                if getattr(self, "transport", None):
                    try:
                        self.transport.close()
                    except Exception:
                        pass
                    # signal on_con_lost if present
                    try:
                        if self.on_con_lost and not self.on_con_lost.is_set():
                            self.on_con_lost.set()
                    except Exception:
                        pass
            elif self.connection_type == "TCP":
                if getattr(self, "writer", None):
                    try:
                        self.writer.close()
                        await self.writer.wait_closed()
                    except Exception:
                        pass
                    self.writer = None
                    self.reader = None
            elif self.connection_type == "BLE":
                # BLE specific cleanup if available
                try:
                    if hasattr(self, "close_ble"):
                        await self.close_ble()
                except Exception:
                    pass
        except Exception:
            pass

        # final cleanup of device tasks
        await self.shutdown()



import logging
from sqlalchemy.orm import aliased
from smartx_rfid.db._main import DatabaseManager
from smartx_rfid.models.orders import ReadersType, Readers, Orders
from datetime import datetime
from smartx_rfid.models.users import Users

connection_string_example = "mysql+pymysql://smartx:smartx@192.168.1.200:3303/smartx"


class SmtxDb:
    def __init__(self, connection_string: str):
        self.connection_string = connection_string
        self.db_manager = DatabaseManager(self.connection_string)
        self.db_manager.initialize()

    # Readers Type
    def get_reader_types(self):
        try:
            with self.db_manager.get_session() as session:
                results = session.query(ReadersType).all()
                return [r.to_dict() for r in results]
        except Exception as e:
            logging.error(f"Error fetching readers type: {e}")
            return []

    def get_reader_type(self, reader_type_id: int):
        try:
            with self.db_manager.get_session() as session:
                result = session.query(ReadersType).filter_by(id=reader_type_id).first()
                return result.to_dict() if result else None
        except Exception as e:
            logging.error(f"Error fetching reader type with id {reader_type_id}: {e}")
            return None

    def add_reader_type(self, name: str, description: str | None = None):
        logging.info(f"Adding reader type: name={name}, description={description}")
        try:
            with self.db_manager.get_session() as session:
                reader_type = ReadersType(name=name, description=description)
                session.add(reader_type)
                session.flush()  # Ensure ID is generated
                return True, reader_type.id
        except Exception as e:
            logging.error(f"Error adding reader type: {e}")
            return False, None

    def update_reader_type(self, reader_type_id: int, name: str | None = None, description: str | None = None):
        logging.info(f"Updating reader type id={reader_type_id}, name={name}, description={description}")
        try:
            with self.db_manager.get_session() as session:
                reader_type = session.query(ReadersType).filter_by(id=reader_type_id).first()
                if not reader_type:
                    return False, f"ReaderType with id {reader_type_id} not found"
                if name is not None:
                    reader_type.name = name
                if description is not None:
                    reader_type.description = description
        except Exception as e:
            logging.error(f"Error updating reader type: {e}")
            return False, str(e)
        return True, None

    def delete_reader_type(self, reader_type_id: int):
        try:
            with self.db_manager.get_session() as session:
                reader_type = session.query(ReadersType).filter_by(id=reader_type_id).first()
                if not reader_type:
                    return False, f"ReaderType with id {reader_type_id} not found"
                session.delete(reader_type)
        except Exception as e:
            logging.error(f"Error deleting reader type: {e}")
            return False, str(e)
        return True, None

    # Readers
    def add_reader(self, reader_type_id: int, serial_number: str, hostname: str | None = None):
        logging.info(
            f"Adding reader: reader_type_id={reader_type_id}, serial_number={serial_number}, hostname={hostname}"
        )
        try:
            with self.db_manager.get_session() as session:
                if not session.query(ReadersType).filter_by(id=reader_type_id).first():
                    return False, f"ReaderType with id {reader_type_id} not found"
                reader = Readers(reader_type_id=reader_type_id, serial_number=serial_number, hostname=hostname)
                session.add(reader)
                session.flush()  # Ensure ID is generated
                return True, reader.id
        except Exception as e:
            logging.error(f"Error adding reader: {e}")
            return False, None

    def get_readers(self, filters: dict | None = None):
        try:
            with self.db_manager.get_session() as session:
                query = session.query(Readers, ReadersType.name.label("reader_type_name")).outerjoin(
                    ReadersType, Readers.reader_type_id == ReadersType.id
                )

                # Apply filters if provided
                if filters:
                    for field, value in filters.items():
                        if hasattr(Readers, field):
                            query = query.filter(getattr(Readers, field) == value)
                        elif hasattr(ReadersType, field):
                            query = query.filter(getattr(ReadersType, field) == value)
                        else:
                            logging.warning(f"Unknown filter field: {field}")

                results = query.all()
                decoded = []
                for reader, reader_type_name in results:
                    d = reader.to_dict()
                    d["reader_type_name"] = reader_type_name
                    decoded.append(d)
                return decoded
        except Exception as e:
            logging.error(f"Error fetching readers: {e}")
            return []

    def get_available_readers(self):
        readers = self.get_readers(filters={"available": True})
        return readers

    def get_readers_by_type(self, reader_type_id: int):
        readers = self.get_readers(filters={"reader_type_id": reader_type_id})
        return readers

    def get_readers_by_type_name(self, type_name: str):
        readers = self.get_readers(filters={"name": type_name})
        return readers

    def get_readers_by_availability(self, available: bool):
        readers = self.get_readers(filters={"available": available})
        return readers

    def get_reader(self, reader_id: int):
        readers = self.get_readers(filters={"id": reader_id})
        return readers[0] if readers else None

    def get_reader_by_hostname(self, hostname: str):
        readers = self.get_readers(filters={"hostname": hostname})
        return readers[0] if readers else None

    def get_reader_by_ids(self, reader_ids: list):
        readers = self.get_readers(filters={"id": reader_ids})
        return readers

    def get_reader_by_serial(self, serial_number: str):
        readers = self.get_readers(filters={"serial_number": serial_number})
        return readers[0] if readers else None

    def update_reader(
        self,
        reader_id: int,
        reader_type_id: int | None = None,
        serial_number: str | None = None,
        hostname: str | None = None,
        available: bool | None = None,
    ):
        logging.info(
            f"Updating reader id={reader_id}, reader_type_id={reader_type_id}, serial_number={serial_number}, hostname={hostname}, available={available}"
        )
        try:
            with self.db_manager.get_session() as session:
                reader = session.query(Readers).filter_by(id=reader_id).first()
                if not reader:
                    return False, f"Reader with id {reader_id} not found"
                if reader_type_id is not None:
                    if not session.query(ReadersType).filter_by(id=reader_type_id).first():
                        return False, f"ReaderType with id {reader_type_id} not found"
                    reader.reader_type_id = reader_type_id
                if serial_number is not None:
                    reader.serial_number = serial_number
                if hostname is not None:
                    reader.hostname = hostname
                if available is not None:
                    reader.available = available
        except Exception as e:
            logging.error(f"Error updating reader: {e}")
            return False, str(e)
        return True, None

    def delete_reader(self, reader_id: int):
        try:
            with self.db_manager.get_session() as session:
                reader = session.query(Readers).filter_by(id=reader_id).first()
                if not reader:
                    return False, f"Reader with id {reader_id} not found"
                in_use = session.query(Orders).filter_by(reader_id=reader_id).first()
                if in_use:
                    return False, f"Reader with id {reader_id} is in use by a product order"
                session.delete(reader)
        except Exception as e:
            logging.error(f"Error deleting reader: {e}")
            return False, str(e)
        return True, None

    # Product Orders
    def get_product_orders(self, filters: dict | None = None):
        """
        Busca pedidos de produto, aceitando filtros de igualdade e de intervalo (__gte, __lte).
        """
        try:
            with self.db_manager.get_session() as session:
                CreatedBy = aliased(Users)
                MountedBy = aliased(Users)
                TestedBy = aliased(Users)
                ShippedBy = aliased(Users)
                ActivatedBy = aliased(Users)

                query = (
                    session.query(
                        Orders,
                        Readers.serial_number.label("reader_serial"),
                        Readers.hostname.label("reader_hostname"),
                        ReadersType.name.label("reader_type_name"),
                        CreatedBy.username.label("created_by_username"),
                        MountedBy.username.label("mounted_by_username"),
                        TestedBy.username.label("tested_by_username"),
                        ShippedBy.username.label("shipped_by_username"),
                        ActivatedBy.username.label("activated_by_username"),
                    )
                    .outerjoin(Readers, Orders.reader_id == Readers.id)
                    .outerjoin(ReadersType, Readers.reader_type_id == ReadersType.id)
                    .outerjoin(CreatedBy, Orders.created_by == CreatedBy.id)
                    .outerjoin(MountedBy, Orders.mounted_by == MountedBy.id)
                    .outerjoin(TestedBy, Orders.tested_by == TestedBy.id)
                    .outerjoin(ShippedBy, Orders.shipped_by == ShippedBy.id)
                    .outerjoin(ActivatedBy, Orders.activated_by == ActivatedBy.id)
                )

                # Aplica filtros de igualdade e de intervalo
                if filters:
                    for field, value in filters.items():
                        if "__gte" in field:
                            base_field = field.replace("__gte", "")
                            if hasattr(Orders, base_field):
                                query = query.filter(getattr(Orders, base_field) >= value)
                        elif "__lte" in field:
                            base_field = field.replace("__lte", "")
                            if hasattr(Orders, base_field):
                                query = query.filter(getattr(Orders, base_field) <= value)
                        elif hasattr(Orders, field):
                            query = query.filter(getattr(Orders, field) == value)
                        elif hasattr(Readers, field):
                            query = query.filter(getattr(Readers, field) == value)
                        elif hasattr(ReadersType, field):
                            query = query.filter(getattr(ReadersType, field) == value)
                        else:
                            logging.warning(f"Unknown filter field: {field}")

                results = query.all()
                decoded = []
                for (
                    order,
                    reader_serial,
                    reader_hostname,
                    reader_type_name,
                    created_by_username,
                    mounted_by_username,
                    tested_by_username,
                    shipped_by_username,
                    activated_by_username,
                ) in results:
                    d = order.to_dict()
                    d["reader_serial"] = reader_serial
                    d["reader_hostname"] = reader_hostname
                    d["reader_type_name"] = reader_type_name
                    d["created_by_username"] = created_by_username
                    d["mounted_by_username"] = mounted_by_username
                    d["tested_by_username"] = tested_by_username
                    d["shipped_by_username"] = shipped_by_username
                    d["activated_by_username"] = activated_by_username
                    decoded.append(d)
                return decoded
        except Exception as e:
            logging.error(f"Error fetching product orders: {e}")
            return []

    def get_product_order(self, order_id: int):
        order = self.get_product_orders(filters={"id": order_id})
        return order[0] if order else None

    def get_product_orders_by_ids(self, order_ids: list):
        orders = self.get_product_orders(filters={"id": order_ids})
        return orders

    def get_product_orders_by_client(self, client: str):
        orders = self.get_product_orders(filters={"client_name": client})
        return orders

    def get_product_orders_by_cnpj(self, cnpj: str):
        orders = self.get_product_orders(filters={"client_cnpj": cnpj})
        return orders

    def get_product_orders_by_product_code(self, product_code: str):
        orders = self.get_product_orders(filters={"product_code": product_code})
        return orders

    def get_product_orders_by_order_number(self, order_number: int):
        orders = self.get_product_orders(filters={"order_number": order_number})
        return orders

    def get_product_orders_by_label_code(self, label_code: str):
        orders = self.get_product_orders(filters={"label_code": label_code})
        return orders

    def get_product_orders_by_date(self, start_date: datetime, end_date: datetime, field: str = "created_at"):
        if getattr(Orders, field, None) is None:
            logging.error(f"Invalid field '{field}' for date filtering in Orders")
            return []
        filters = {f"{field}__gte": start_date, f"{field}__lte": end_date}
        try:
            return self.get_product_orders(filters)
        except Exception as e:
            logging.error(f"Error fetching product orders between {start_date} and {end_date}: {e}")
            return []

    def get_product_orders_by_reader_type(self, reader_type_id: int):
        return self.get_product_orders(filters={"reader_type_id": reader_type_id})

    def get_product_orders_by_reader_type_name(self, reader_type_name: str):
        type_id = self.get_reader_types()
        type_id = next((t.get("id") for t in type_id if t.get("name") == reader_type_name), None)
        if type_id is None:
            return []
        return self.get_product_orders(filters={"reader_type_id": type_id})

    def get_product_orders_by_reader(self, reader_id: int):
        orders = self.get_product_orders(filters={"reader_id": reader_id})
        return orders

    def add_product_order(
        self,
        order_number: int,
        client_name: str,
        client_cnpj: str | None,
        product_code: str,
        product_description: str | None,
        product_family: str | None,
        reader_id: int | None = None,
        created_by: int | None = None,
    ):
        try:
            with self.db_manager.get_session() as session:
                # Validate created_by user exists
                if created_by is not None:
                    if not session.query(Users).filter_by(id=created_by).first():
                        return False, f"User with id {created_by} not found"

                # Validate reader_id and check availability
                if reader_id is not None:
                    reader = session.query(Readers).filter_by(id=reader_id).first()
                    if not reader:
                        return False, f"Reader with id {reader_id} not found"
                    if not reader.available:
                        return False, f"Reader with id {reader_id} is not available"

                product_order = Orders(
                    order_number=order_number,
                    client_name=client_name,
                    client_cnpj=client_cnpj,
                    product_code=product_code,
                    product_description=product_description,
                    product_family=product_family,
                    reader_id=reader_id,
                    created_by=created_by,
                )
                session.add(product_order)
                if reader_id is not None:
                    reader.available = False
                session.flush()  # Ensure ID is generated
                return True, product_order.id
        except Exception as e:
            logging.error(f"Error adding product order: {e}")
            return False, str(e)

    def update_product_order(self, order_id: int, **kwargs):
        logging.info(f"Updating product order id={order_id}, updates={kwargs}")
        try:
            with self.db_manager.get_session() as session:
                product_order = session.query(Orders).filter_by(id=order_id).first()
                if not product_order:
                    return False, f"Orders with id {order_id} not found"
                for key, value in kwargs.items():
                    if hasattr(product_order, key):
                        setattr(product_order, key, value)
        except Exception as e:
            logging.error(f"Error updating product order: {e}")
            return False, str(e)
        return True, None

    def delete_product_order(self, order_id: int):
        try:
            with self.db_manager.get_session() as session:
                product_order = session.query(Orders).filter_by(id=order_id).first()
                if not product_order:
                    return False, f"Orders with id {order_id} not found"
                # Restore reader availability
                reader = session.query(Readers).filter_by(id=product_order.reader_id).first()
                if reader:
                    reader.available = True
                session.delete(product_order)
        except Exception as e:
            logging.error(f"Error deleting product order: {e}")
            return False, str(e)
        return True, None

    def add_reader_to_product_order(self, order_id: int, reader_id: int):
        logging.info(f"Adding reader id={reader_id} to product order id={order_id}")
        try:
            with self.db_manager.get_session() as session:
                product_order = session.query(Orders).filter_by(id=order_id).first()
                if not product_order:
                    return False, f"Orders with id {order_id} not found"

                reader = session.query(Readers).filter_by(id=reader_id).first()
                if not reader:
                    return False, f"Reader with id {reader_id} not found"
                if not reader.available:
                    return False, f"Reader with id {reader_id} is not available"

                # Free previous reader if any
                if product_order.reader_id:
                    old_reader = session.query(Readers).filter_by(id=product_order.reader_id).first()
                    if old_reader:
                        old_reader.available = True

                product_order.reader_id = reader_id
                reader.available = False
        except Exception as e:
            logging.error(f"Error adding reader to product order: {e}")
            return False, str(e)
        return True, None

    def add_comment_to_product_order(self, order_id: int, comment: str, user: str | None = None):
        if user is None:
            user = "Unknown"
        comment = f"{datetime.now().isoformat()} - {user}: {comment}"
        logging.info(f"Adding comment to product order id={order_id}: {comment}")
        try:
            with self.db_manager.get_session() as session:
                product_order = session.query(Orders).filter_by(id=order_id).first()
                if not product_order:
                    return False, f"Orders with id {order_id} not found"
                existing_comments = product_order.comments or ""
                updated_comments = existing_comments + "\n" + comment if existing_comments else comment
                product_order.comments = updated_comments
        except Exception as e:
            logging.error(f"Error adding comment to product order: {e}")
            return False, str(e)
        return True, comment

    # [ Product Orders Workflow ]
    def product_order_mount(self, order_id: int, mounted_by: int):
        order = self.get_product_order(order_id)
        if not order:
            return False, f"Orders with id {order_id} not found"
        if order.get("reader_id") is None:
            return False, "Order must have a reader assigned before mounting"
        if order.get("mounted_at") is not None:
            return False, "Order already mounted"
        return self.update_product_order(order_id, mounted_at=datetime.now(), mounted_by=mounted_by)

    def product_order_test(self, order_id: int, tested_by: int):
        order = self.get_product_order(order_id)
        if not order:
            return False, f"Orders with id {order_id} not found"
        if order.get("mounted_at") is None:
            return False, "Order must be mounted before testing"
        if order.get("tested_at") is not None:
            return False, "Order already tested"
        return self.update_product_order(order_id, tested_at=datetime.now(), tested_by=tested_by)

    def product_order_ship(self, order_id: int, shipped_by: int):
        order = self.get_product_order(order_id)
        if not order:
            return False, f"Orders with id {order_id} not found"
        if order.get("tested_at") is None:
            return False, "Order must be tested before shipping"
        if order.get("shipped_at") is not None:
            return False, "Order already shipped"
        return self.update_product_order(order_id, shipped_at=datetime.now(), shipped_by=shipped_by)

    def product_order_activate(self, order_id: int, activated_by: int):
        order = self.get_product_order(order_id)
        if not order:
            return False, f"Orders with id {order_id} not found"
        if order.get("shipped_at") is None:
            return False, "Order must be shipped before activation"
        if order.get("activated_at") is not None:
            return False, "Order already activated"
        return self.update_product_order(order_id, activated_at=datetime.now(), activated_by=activated_by)

    # [ USERS ]
    def get_users(self):
        try:
            with self.db_manager.get_session() as session:
                results = session.query(Users).all()
                return [r.to_dict() for r in results]
        except Exception as e:
            logging.error(f"Error fetching users: {e}")
            return []

    def get_user(self, user_id: int):
        try:
            with self.db_manager.get_session() as session:
                result = session.query(Users).filter_by(id=user_id).first()
                return result.to_dict() if result else None
        except Exception as e:
            logging.error(f"Error fetching user with id {user_id}: {e}")
            return None

    def get_user_by_username(self, username: str):
        try:
            with self.db_manager.get_session() as session:
                result = session.query(Users).filter_by(username=username).first()
                return result.to_dict() if result else None
        except Exception as e:
            logging.error(f"Error fetching user with username {username}: {e}")
            return None

    def add_user(self, username: str, password_hash: str, role: str = "user"):
        logging.info(f"Adding user: username={username}, role={role}")
        try:
            with self.db_manager.get_session() as session:
                if session.query(Users).filter_by(username=username).first():
                    return False, f"User with username {username} already exists"
                user = Users(username=username, password_hash=password_hash, role=role)
                session.add(user)
                session.flush()  # Ensure ID is generated
                return True, user.id
        except Exception as e:
            logging.error(f"Error adding user: {e}")
            return False, None

    def update_user(self, user_id: int, **kwargs):
        logging.info(f"Updating user id={user_id}, updates={kwargs}")
        try:
            with self.db_manager.get_session() as session:
                user = session.query(Users).filter_by(id=user_id).first()
                if not user:
                    return False, f"User with id {user_id} not found"
                for key, value in kwargs.items():
                    if hasattr(user, key):
                        setattr(user, key, value)
        except Exception as e:
            logging.error(f"Error updating user: {e}")
            return False, str(e)
        return True, None

    def delete_user(self, user_id: int):
        try:
            with self.db_manager.get_session() as session:
                user = session.query(Users).filter_by(id=user_id).first()
                if not user:
                    return False, f"User with id {user_id} not found"
                session.delete(user)
        except Exception as e:
            logging.error(f"Error deleting user: {e}")
            return False, str(e)
        return True, None

    def delete_user_by_username(self, username: str):
        try:
            with self.db_manager.get_session() as session:
                user = session.query(Users).filter_by(username=username).first()
                if not user:
                    return False, f"User with username {username} not found"
                session.delete(user)
        except Exception as e:
            logging.error(f"Error deleting user by username {username}: {e}")
            return False, str(e)
        return True, None

    # [ UTILS ]
    def _get_column_values(self, model, column_name):
        try:
            with self.db_manager.get_session() as session:
                column = getattr(model, column_name, None)
                if column is None:
                    logging.error(f"Model {model.__name__} does not have column {column_name}")
                    return []
                results = session.query(column).distinct().all()
                return [getattr(r, column_name) for r in results if getattr(r, column_name) is not None]
        except Exception as e:
            logging.error(f"Error fetching column values for {model.__name__}.{column_name}: {e}")
            return []

    def get_customers(self):
        return self._get_column_values(Orders, "client_name")

    def get_cnpjs(self):
        return self._get_column_values(Orders, "client_cnpj")

    def get_orders_numbers(self):
        return self._get_column_values(Orders, "order_number")

    def get_product_codes(self):
        return self._get_column_values(Orders, "product_code")



import json
import base64
import hashlib
import logging
import uuid
from datetime import datetime, timedelta
from typing import Optional, Dict, Any

from cryptography.hazmat.primitives import hashes, serialization
from cryptography.hazmat.primitives.asymmetric import rsa, padding
from cryptography.exceptions import InvalidSignature


class LicenseManager:
    """
    Professional License Manager

    Features:
    - RSA signing
    - Hardware binding
    - Expiration support
    - Dict-like access
    - License request string (for activation systems)
    """

    # ==========================================================
    # INIT
    # ==========================================================
    def __init__(self, private_key_pem: Optional[str] = None, public_key_pem: Optional[str] = None):
        self.private_key = None
        self.public_key = None
        self.public_key_pem: Optional[str] = None
        self.license_data: Optional[Dict[str, Any]] = None

        if private_key_pem:
            self.load_private_key(private_key_pem)

        if public_key_pem:
            self.load_public_key(public_key_pem)

    # ==========================================================
    # LICENSE REQUEST STRING
    # ==========================================================
    def build_license_request_string(self, public_key_pem: str | None = None, hardware_id: Optional[str] = None) -> str:
        """
        Returns a base64 string containing hardware_id and public_key.
        Useful for license activation systems.
        """
        if not self.public_key_pem and not public_key_pem:
            raise Exception("Public key must be provided either in constructor or as argument")
        if public_key_pem is None:
            public_key_pem = self.public_key_pem
        if hardware_id is None:
            hardware_id = LicenseManager.get_hardware_id()

        payload = json.dumps({"hardware_id": hardware_id, "public_key": public_key_pem}, sort_keys=True).encode()

        return base64.b64encode(payload).decode()

    @staticmethod
    def parse_license_request_string(request_string: str) -> Dict[str, str]:
        """
        Decodes a license request string.
        Returns dict with hardware_id and public_key.
        """

        try:
            payload = base64.b64decode(request_string)
            return json.loads(payload.decode())
        except Exception as e:
            raise ValueError(f"Invalid license request string: {e}")

    # ==========================================================
    # KEY MANAGEMENT
    # ==========================================================
    @staticmethod
    def generate_key_pair(key_size: int = 2048):
        """
        Generate RSA key pair.
        Returns (private_pem, public_pem)
        """

        private_key = rsa.generate_private_key(public_exponent=65537, key_size=key_size)

        private_pem = private_key.private_bytes(
            encoding=serialization.Encoding.PEM,
            format=serialization.PrivateFormat.PKCS8,
            encryption_algorithm=serialization.NoEncryption(),
        ).decode()

        public_pem = (
            private_key.public_key()
            .public_bytes(encoding=serialization.Encoding.PEM, format=serialization.PublicFormat.SubjectPublicKeyInfo)
            .decode()
        )

        return private_pem.strip(), public_pem.strip()

    def load_private_key(self, private_key_pem: str):
        self.private_key = serialization.load_pem_private_key(private_key_pem.encode(), password=None)

    def load_public_key(self, public_key_pem: str):
        self.public_key = serialization.load_pem_public_key(public_key_pem.encode())
        self.public_key_pem = public_key_pem

    # ==========================================================
    # HARDWARE ID
    # ==========================================================
    @staticmethod
    def get_hardware_id() -> str:
        mac = uuid.getnode()
        raw = str(mac).encode()
        return hashlib.sha256(raw).hexdigest()

    # ==========================================================
    # LICENSE CREATION
    # ==========================================================
    def create_license(
        self, data: Dict[str, Any], duration_days: Optional[int] = None, hardware_id: Optional[str] = None
    ) -> str:
        """
        Create and sign license.
        Requires private key.
        """

        if not self.private_key:
            raise Exception("Private key not loaded")

        license_data = data.copy()

        # Hardware binding
        if hardware_id is None:
            hardware_id = self.get_hardware_id()

        license_data["hardware_id"] = hardware_id

        # Expiration
        if duration_days is not None:
            expires = datetime.now() + timedelta(days=duration_days)
            license_data["expires"] = expires.isoformat()
        elif "expires" not in license_data:
            expires = datetime.now() + timedelta(days=365)
            license_data["expires"] = expires.isoformat()

        payload = json.dumps(license_data, sort_keys=True).encode()

        signature = self.private_key.sign(payload, padding.PKCS1v15(), hashes.SHA256())

        combined = payload + b"||" + signature
        return base64.b64encode(combined).decode()

    # ==========================================================
    # LICENSE VALIDATION
    # ==========================================================
    def load_license(self, license_string: str):
        if not self.public_key:
            raise Exception("Public key not loaded")

        decoded = base64.b64decode(license_string)
        payload, signature = decoded.split(b"||")

        # Verify signature
        try:
            self.public_key.verify(signature, payload, padding.PKCS1v15(), hashes.SHA256())
        except InvalidSignature:
            raise Exception("Invalid license signature")

        data = json.loads(payload.decode())

        # Hardware validation
        if data.get("hardware_id") != self.get_hardware_id():
            raise Exception("License not valid for this hardware")

        # Expiration validation
        if data.get("expires"):
            exp = datetime.fromisoformat(data["expires"])
            if datetime.now() > exp:
                raise Exception("License expired")

        self.license_data = data
        logging.info("License loaded and valid")

    def validate_license(self) -> bool:
        if not self.license_data:
            return False
        try:
            expires = self.license_data.get("expires")
            if expires:
                exp = datetime.fromisoformat(expires)
                if datetime.now() > exp:
                    return False
            return True
        except Exception:
            return False

    # ==========================================================
    # DICT-LIKE ACCESS
    # ==========================================================
    def get(self, key, default=None):
        if not self.license_data:
            return default
        return self.license_data.get(key, default)

    def __getitem__(self, key):
        if not self.license_data:
            raise Exception("License not loaded")
        return self.license_data[key]

    def __contains__(self, key):
        if not self.license_data:
            return False
        return key in self.license_data

    # ==========================================================
    # UTIL
    # ==========================================================
    def is_loaded(self) -> bool:
        return self.license_data is not None



import asyncio
from typing import Callable, Any, Awaitable, Union


async def delayed_function(func: Union[Callable, Awaitable], delay: float, *args, **kwargs) -> Any:
    """
    Executes a function or coroutine after a specified delay, always asynchronously.
    :param func: Synchronous function or coroutine to be executed
    :param delay: Delay time in seconds
    :param args: Positional arguments for the function
    :param kwargs: Keyword arguments for the function
    :return: The result of the function or coroutine
    """
    await asyncio.sleep(delay)
    if asyncio.iscoroutinefunction(func):
        return await func(*args, **kwargs)
    else:
        return func(*args, **kwargs)



def generate_zpl_with_params(zpl: str, **kwargs) -> str:
    """
    Replaces ZPL placeholders (e.g., {epc}, {num}) with values provided in kwargs.
    Example:
            generate_zpl_with_params(zpl, epc="123", num="456")
    """
    try:
        return zpl.format(**kwargs)
    except KeyError as e:
        raise ValueError(f"Missing parameter for placeholder: {e}")



import asyncio
import logging
from typing import Callable, Optional

from smartx_rfid.devices._base import DeviceBase
from smartx_rfid.utils.event import on_event
from .helpers import Helpers
from smartx_rfid.utils import get_hash


class SatoPrinter(DeviceBase, Helpers):
    def __init__(
        self,
        ip: str,
        name: str = "Sato",
        port: int = 9100,
        reconnection_time: int = 3,
        **kwargs,
    ):
        DeviceBase.__init__(self)
        self.name = name
        self.device_type = "printer"
        self.ip = ip
        self.port = port
        self.reconnection_time = reconnection_time
        self.is_connected: bool = False
        self.on_event: Callable = on_event
        self._reader: Optional[asyncio.StreamReader] = None
        self._writer: Optional[asyncio.StreamWriter] = None
        self._connect_lock = asyncio.Lock()
        self._running = True

        self.name = name

        self.ip = ip
        self.port = port

        self.reader = None
        self.writer = None

        self.is_connected = False
        self.can_print = False
        self.last_status = None
        self.on_event: Callable = on_event

        self._to_print: list[str] = []

        # Success on print
        self._print_sent = False
        self._zpl_id = None

    async def connect(self):
        """Connect to TCP server and keep connection alive."""
        while self._running:
            try:
                logging.info(f"Connecting: {self.name} - {self.ip}:{self.port}")
                self.reader, self.writer = await asyncio.wait_for(
                    asyncio.open_connection(self.ip, self.port), timeout=3
                )
                if not self.is_connected:
                    self.is_connected = True
                    self.on_event(self.name, "connection", True)

                # Start the receive and monitor tasks
                tasks = [
                    self.create_task(self.receive_data()),
                    self.create_task(self.get_status()),
                ]

                # Wait until one of the tasks completes (e.g. disconnection)
                done, pending = await asyncio.wait(tasks, return_when=asyncio.FIRST_COMPLETED)

                # Cancel any remaining tasks
                for task in pending:
                    task.cancel()

                self.is_connected = False
                self.on_event(self.name, "connection", False)

            except Exception as e:
                if self.is_connected:
                    self.is_connected = False
                    self.on_event(self.name, "connection", False)
                logging.error(f"[CONNECTION ERROR] {e}")

            await asyncio.sleep(3)

    async def close(self):
        # stop connect loop
        self._running = False

        # close writer if present
        try:
            if self.writer:
                try:
                    self.writer.close()
                    await self.writer.wait_closed()
                except Exception:
                    pass
                self.writer = None
                self.reader = None
        except Exception:
            pass

        await self.shutdown()

    async def write(self, data: str, verbose=True):
        """
        Send data through TCP connection.

        Args:
            data: Text to send
            verbose: Show sent data in logs
        """
        if self.is_connected and self.writer:
            try:
                to_send = (data + "\n").encode("utf-8")
                self.writer.write(to_send)
                await self.writer.drain()
                if verbose:
                    logging.info(f"[{self.name}] [SENT] {data.strip()}")
            except Exception as e:
                logging.warning(f"[{self.name}] [SEND ERROR] data={data.strip()} error={e}")
                if self.is_connected:
                    self.is_connected = False
                    self.on_event(self.name, "connection", False)

    def print(self, zpl: str):
        """Send ZPL command to printer."""
        if not self.is_connected:
            return False, "Printer not connected."
        if not self.can_print:
            return False, "Printer not ready."
        asyncio.create_task(self.write(zpl))
        self._print_sent = True
        self._zpl_id = get_hash(zpl)
        return True, self._zpl_id

    def add_to_print_queue(self, zpl: str | list[str]):
        """Add ZPL command to the print queue."""
        if isinstance(zpl, list):
            self._to_print = self._to_print + zpl
        else:
            self._to_print.append(zpl)
        logging.info(f"{self.name} - Added to print queue. {len(self._to_print)} total.")



class AlertsManager:
    def __init__(self):
        self.alerts = []

    def get_alerts(self):
        alerts = self.alerts.copy()
        self.alerts.clear()  # Clear after fetching
        return alerts

    def add_alert(self, message: str, level: str = "info"):
        alert = {"message": message, "level": level}
        self.alerts.append(alert)

    def add_info(self, message: str):
        self.add_alert(message, "info")

    def add_warning(self, message: str):
        self.add_alert(message, "warning")

    def add_error(self, message: str):
        self.add_alert(message, "error")

    def add_success(self, message: str):
        self.add_alert(message, "success")



# GENERIC
from .generic.SERIAL._main import SERIAL
from .generic.TCP._main import TCP

# RFID DEVICES
from .RFID.X714._main import X714
from .RFID.R700_IOT._main import R700_IOT
from .RFID.R700_IOT.reader_config_example import R700_IOT_config_example
from .RFID.ACUPAD._main import ACUPAD

# PRINTER DEVICES
from .printer.SATO._main import SatoPrinter

# Device Manager
from .device_manager import DeviceManager



EPC_LENGTH = 24


def data_to_hex(data: str, pad_char: str = " ") -> str:
    return "".join([f"{ord(c):02X}" for c in data.rjust(12, pad_char)])


def parse_encode_rule(params: str) -> dict:
    """Parse encode rule string (key=value\\r\\n pairs) into a dict."""
    rule = {}
    for line in params.replace("\r\n", "\n").replace("\r", "\n").split("\n"):
        if "=" in line:
            key, _, value = line.partition("=")
            rule[key.strip()] = value.strip()
    return rule


def build_epc(serial: int, rule: dict) -> str:
    """Build a single EPC string from a serial number and encode rule.

    The result is always exactly 24 hex characters (96-bit EPC).
    """
    # prefix = rule.get("EPCPrefix", "") or ""
    # suffix = rule.get("EPCSuffix", "") or ""
    # pad_char = rule.get("PadChar", "0") or "0"
    # pad_align = rule.get("PadAlign", "LEFT").upper()
    # barcode_prefix_len = int(rule.get("BarcodePrefixLength", 0) or 0)
    # barcode_suffix_len = int(rule.get("BarcodeSuffixLength", 0) or 0)
    # convert_to_hex = (rule.get("ConvertToHex", "ALL") or "ALL").upper()

    return None



"""
RFID models for SMARTX Connector.

Defines the Tag and Event models for storing RFID reader data
with proper indexing and relationships.
"""

from sqlalchemy import Column, String, Text, Integer

from .mixin import Base, BaseMixin


class License(Base, BaseMixin):
    __tablename__ = "licenses"
    id = Column(Integer, primary_key=True, autoincrement=True)
    public_key = Column(String(2000), nullable=False, index=True, unique=True)
    private_key = Column(Text, nullable=False)

    def __setattr__(self, name, value):
        if name in ["public_key", "private_key"] and isinstance(value, str):
            value = value.strip()
        super().__setattr__(name, value)
